﻿using System;
using System.Linq;
using System.Security.Claims;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.IdentityModel.Tokens;
using Tsi1.Api.Infrastructure;
using Tsi1.Api.Models;
using Tsi1.BusinessLayer.Dtos;
using Tsi1.BusinessLayer.Helpers;
using Tsi1.BusinessLayer.Interfaces;

namespace Tsi1.Api.Controllers
{
    [Authorize]
    [Route("api/[controller]")]
    [ApiController]
    public class UserController : ControllerBase
    {
        private readonly IJwtAuthManager _jwtAuthManager;
        private readonly IUserService _userService;
        private readonly IUserTypeService _userTypeService;
        private readonly ITenantService _tenantService;

        public UserController(IJwtAuthManager jwtAuthManager, IUserService userService,
            IUserTypeService userTypeService, ITenantService tenantService)
        {
            _jwtAuthManager = jwtAuthManager;
            _userService = userService;
            _userTypeService = userTypeService;
            _tenantService = tenantService;
        }

        [AllowAnonymous]
        [HttpPost("Login")]
        public async Task<IActionResult> Login(LoginRequest request)
        {
            var resultSplit = request.UserName.Split("@");

            if (resultSplit.Count() != 2)
            {
                return BadRequest(ErrorMessages.InvalidUsername);
            }

            var userName = resultSplit[0];

            var tenantName = resultSplit[1];

            var tenantId = await _tenantService.GetByName(tenantName);

            if (tenantId.HasError)
            {
                return BadRequest(tenantId.Message);
            }

            var result = await _userService.Authenticate(userName, request.Password, tenantId.Data);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            var user = result.Data;

            var claims = new[]
            {
                new Claim("Id", user.Id.ToString()),
                new Claim("Username", user.Username),
                new Claim("TenantId", user.TenantId.ToString()),
                new Claim(ClaimTypes.Role, user.UserType.Name)
            };

            var jwtResult = _jwtAuthManager.GenerateTokens(user.Username, claims, DateTime.Now);

            return Ok(new LoginResult
            {
                UserName = user.Username,
                Role = user.UserType.Name,
                AccessToken = jwtResult.AccessToken,
                RefreshToken = jwtResult.RefreshToken.TokenString
            });
        }

        [AllowAnonymous]
        [HttpPost("RefreshToken")]
        public async Task<ActionResult> RefreshToken([FromBody] RefreshTokenRequest request)
        {
            try
            {
                var username = HttpContext.User.Claims.FirstOrDefault(x => x.Type == "Username").Value;

                if (string.IsNullOrWhiteSpace(request.RefreshToken))
                {
                    return Unauthorized();
                }

                var accessToken = await HttpContext.GetTokenAsync("Bearer", "access_token");
                var jwtResult = _jwtAuthManager.Refresh(request.RefreshToken, accessToken, DateTime.Now);

                return Ok(new LoginResult
                {
                    UserName = username,
                    Role = User.FindFirst(ClaimTypes.Role)?.Value ?? string.Empty,
                    AccessToken = jwtResult.AccessToken,
                    RefreshToken = jwtResult.RefreshToken.TokenString
                });
            }
            catch (SecurityTokenException e)
            {
                return Unauthorized(e.Message); // return 401 so that the client side can redirect the user to login page
            }
        }

        [Authorize(Roles = UserTypes.FacultyAdmin)]
        [HttpPost("Register")]
        public async Task<IActionResult> Register(UserRegisterDto dto)
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

            var userTypeResult = await _userTypeService.GetById(dto.UserTypeId);

            if (userTypeResult.HasError)
            {
                return BadRequest(userTypeResult.Message);
            }

            var userType = userTypeResult.Data;

            if (userType.Name == UserTypes.UdelarAdmin ||
                userType.Name == UserTypes.FacultyAdmin)
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            var userServiceResult = await _userService.Create(dto, userType.Name, tenantId);

            if (userServiceResult.HasError)
            {
                return BadRequest(userServiceResult.Message);
            }

            return Ok();
        }

        [Authorize(Roles = UserTypes.UdelarAdmin)]
        [HttpPost("RegisterAdmin/{tenantId}")]
        public async Task<IActionResult> RegisterAdmin(UserRegisterDto dto, int tenantId)
        {
            var userTypeResult = await _userTypeService.GetById(dto.UserTypeId);

            if (userTypeResult.HasError)
            {
                BadRequest(userTypeResult.Message);
            }

            var userType = userTypeResult.Data;

            if (userType.Name == UserTypes.Student ||
                userType.Name == UserTypes.Professor)
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            var userServiceResult = await _userService.Create(dto, userType.Name, tenantId);

            if (userServiceResult.HasError)
            {
                BadRequest(userServiceResult.Message);
            }

            return Ok();
        }


        [Authorize(Roles = UserTypes.Student + ", " + UserTypes.Professor)]
        [HttpGet("GetAll")]
        public async Task<IActionResult> GetAll()
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

            var result = await _userService.GetAll(tenantId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.Student + ", " + UserTypes.Professor)]
        [HttpGet("GetById/{userId}")]
        public async Task<IActionResult> GetById(int userId)
        {
            var result = await _userService.GetById(userId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.UdelarAdmin + ", " + UserTypes.FacultyAdmin)]
        [HttpGet("GetUserTypes")]
        public async Task<IActionResult> GetUserTypes()
        {
            var userType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            var result = await _userTypeService.GetAll(userType);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

    }
}
